package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.logger.getLogger
import jakarta.annotation.PreDestroy
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import org.springframework.stereotype.Component

enum class TaskType{
    REPO_CHANGED,
    DONE
}
data class Task(val type: TaskType, val payload: Any)

/**
 * The Task Manager takes tasks and distributes them to
 * underlying workers. Internally it uses a channel
 * to manage incoming tasks.
 */
@Component
class TaskManager : Worker("Task Manager") {

    private val gitWorker = GitWorker(tasks)

    /**
     * Used to distribute incoming tasks to the fitting workers
     * depending on the task's type.
     */
    override suspend fun executeTask(task: Task) {
        when (task.type) {
            TaskType.REPO_CHANGED -> {
                logger.info("[${Thread.currentThread().name}] add task called in ${task.payload}")
                gitWorker.addTask(task)
                logger.info("[${Thread.currentThread().name}] add task finished in ${task.payload}")
            }
            TaskType.DONE -> {
                logger.info("${Thread.currentThread().name}] Task done ${task.payload}")
            }
        }
    }

    @PreDestroy
    private fun destruct() {
        gitWorker.close()
        close()
    }
}

class GitWorker(private val responseChannel: Channel<Task>): Worker("Git Worker") {

    /**
     * Performs git related tasks, like cloning the given repository.
     */
    override suspend fun executeTask(task: Task) {
        // simulate a long running task
        logger.info("[${Thread.currentThread().name}] Starting task ${task.payload} in gitworker")
        delay(5000)
        responseChannel.send(Task(TaskType.DONE, "Task with payload ${task.payload} finished"))
        logger.info("[${Thread.currentThread().name}] Finished task ${task.payload} in gitworker")
    }

}

/**
 * Abstract worker class used as a base for multithreaded worker.
 * It uses a fixed thread pool as a base for the coroutineScope.
 * The size of the thread pool is equal to the number of cores.
 */
abstract class Worker(private val name: String): IWorker {

    // TODO: this could be dangerous to give it unlimited memory
    protected val tasks = Channel<Task>(Channel.UNLIMITED)
    private val numCores = Runtime.getRuntime().availableProcessors()
    // use a fixed size thread pool context with numCores threads
    @OptIn(DelicateCoroutinesApi::class)
    private val threadPoolContext = newFixedThreadPoolContext(numCores, name)
    private val coroutineScope = CoroutineScope(threadPoolContext)

    protected val logger = getLogger(javaClass)

    override suspend fun addTask(task: Task) {
        tasks.send(task)
    }

    init {
        logger.info("Starting $numCores workers.")
        repeat(numCores) {
            logger.info("[${Thread.currentThread().name}] launching coroutine in $name")
            launchWorker(it)
        }
    }

    /**
     * Wrapper function to execute tasks provided through the channel.
     * We use a dedicated number of coroutines to watch the tasks channel.
     * Each coroutine starts a new coroutine to execute the task. This is
     * done so that the outer coroutines are not blocked by the potentially
     * very long-running inner task performed by executeTask.
     */
    private fun launchWorker(id: Int) = coroutineScope.launch {
        for (task in tasks) {
            coroutineScope.launch {
                logger.debug("[${Thread.currentThread().name}] Processor #$id-$name received $task")
                executeTask(task)
                logger.debug("[${Thread.currentThread().name}] Processor #$id-$name finished $task")
            }
        }
    }

    override fun close() {
        coroutineScope.launch {
            tasks.close()
            for (job in coroutineContext[Job]!!.children) {
                job.join()
            }
            threadPoolContext.close()
            logger.info("Worker $name has been closed")
        }
    }
}

interface IWorker {
    /**
     * Adds a new task to the worker's internal tasks channel.
     */
    suspend fun addTask(task: Task)

    /**
     * Worker dependent function describing the worker's
     * handling of the task.
     * This function is automatically called.
     */
    suspend fun executeTask(task: Task)

    /**
     * Function used to close the underlying resources.
     */
    fun close()
}